﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace ConfigEditor
{
    public partial class ConfigView : UserControl
    {
        public ConfigView()
        {
            InitializeComponent();
        }

        private void btnOpenConfig_Click(object sender, EventArgs e)
        {
            var ofd = new OpenFileDialog()
            {
                DefaultExt = "*.exp",
                Filter = "*.exp|*.exp",
                FilterIndex = 0
            };
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                using (TextReader txt = new StreamReader(ofd.FileName, Encoding.Default))
                {
                    tvConfig.BeginUpdate();
                    tvConfig.Nodes.Clear();
                    while (txt.ReadLine() != "PLC_CONFIGURATION") { }
                    var root = ReadNode(txt, "PLC_END", "PLC_CONFIGURATION");
                    foreach (TreeNode node in root.Nodes)
                    {
                        tvConfig.Nodes.Add(node);
                    }
                    tvConfig.EndUpdate();
                }
            }
        }

        TreeNode ReadNode(TextReader txt, string stopTag, string startTag = "Node")
        {
            var node = new TreeNode()
            {
                Text = startTag,
            };
            var sb = new StringBuilder();
            string line, name;
            while ((line = txt.ReadLine()) != null && (name = GetName(line)) != stopTag)
            {
                switch (name)
                {
                    case "_GLOBAL":
                        node.Nodes.Add(ReadNode(txt, "_END_GLOBAL", line));
                        continue;
                    case "_MODULE":
                        node.Nodes.Add(ReadNode(txt, "_END_MODULE", line));
                        continue;
                    case "_CHANNEL":
                        node.Nodes.Add(ReadNode(txt, "_END_CHANNEL", line));
                        continue;
                    case "_PARAMETER":
                        node.Nodes.Add(ReadNode(txt, "_END_PARAMETER", line));
                        continue;
                }
                if (name == "_MODULE_NAME" || name == "_SYMBOLIC_NAME")
                {
                    node.Text = line.Split(new string[] { ": " }, StringSplitOptions.RemoveEmptyEntries)[1];
                }
                else if (name == "_COMMENT")
                {
                    var comment = line.Split(new string[] { ": " }, StringSplitOptions.RemoveEmptyEntries)[1];
                    if (comment != "''")
                    {
                        node.Text += " - " + comment;
                    }
                }
                if (line != string.Empty)
                {
                    sb.AppendLine(line);
                }
            }
            node.Tag = new Tuple<string, string, string>(startTag, stopTag, sb.ToString().TrimEnd('\r', '\n'));
            return node;
        }

        string GetName(string line)
        {
            if (line != null || line != string.Empty)
            {
                return line.Split(new char[] { ':' })[0];
            }
            else
            {
                return null;
            }
        }

        private void btnSaveConfigAs_Click(object sender, EventArgs e)
        {
            var sfd = new SaveFileDialog()
            {
                DefaultExt = "*.exp",
                Filter = "*.exp|*.exp",
                FilterIndex = 0
            };
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                using (TextWriter txt = new StreamWriter(sfd.FileName, false, Encoding.Default))
                {
                    txt.WriteLine("PLC_CONFIGURATION");
                    foreach (var node in tvConfig.Nodes)
                    {
                        WriteNode(node as TreeNode, txt);
                    }
                    txt.WriteLine("PLC_END");
                }
            }
        }

        private void WriteNode(TreeNode node, TextWriter tw)
        {
            var tuple = node.Tag as Tuple<string, string, string>;
            tw.WriteLine(tuple.Item1);
            tw.WriteLine(tuple.Item3);
            foreach (TreeNode child in node.Nodes)
            {
                WriteNode(child, tw);
            }
            tw.WriteLine(tuple.Item2);
            tw.WriteLine();
        }

        private void tvConfig_AfterSelect(object sender, TreeViewEventArgs e)
        {
            txtProps.Text = (e.Node.Tag as Tuple<string, string, string>).Item3;
        }

        private void tvConfig_DragEnter(object sender, DragEventArgs e)
        {
            e.Effect = DragDropEffects.Move;
        }

        private void tvConfig_DragOver(object sender, DragEventArgs e)
        {
            var clientPos = tvConfig.PointToClient(Cursor.Position);
            var target = tvConfig.GetNodeAt(clientPos);
            var payload = e.Data.GetData(typeof(TreeNode)) as TreeNode;
            if (target == null || target.IsChildOf(payload))
            {
                tvConfig.Refresh();
                return;
            }
            var offset = clientPos.Y - target.Bounds.Top - target.Bounds.Height / 2;
            tvConfig.Refresh();
            // negative - above, zero - inside, positive - below
            DrawPlaceholder(target, offset / (target.Bounds.Height / 3));
        }

        private void tvConfig_DragDrop(object sender, DragEventArgs e)
        {
            var clientPos = tvConfig.PointToClient(Cursor.Position);
            var target = tvConfig.GetNodeAt(clientPos);
            var payload = e.Data.GetData(typeof(TreeNode)) as TreeNode;
            if (target == null || target.IsChildOf(payload))
            {
                tvConfig.Refresh();
                return;
            }
            var offset = clientPos.Y - target.Bounds.Top - target.Bounds.Height / 2;
            tvConfig.Refresh();
            // negative - above, zero - inside, positive - below
            var where = offset / (target.Bounds.Height / 3); // i.e. offset / dead_zone_size
            if (where == 0)
            {
                target.Nodes.Add(payload.Clone() as TreeNode);
            }
            else
            {
                var collection = target.Parent != null ? target.Parent.Nodes : tvConfig.Nodes;
                collection.Insert(target.Index + (where > 0 ? 1 : 0), payload.Clone() as TreeNode);
            }
            payload.Remove();
        }

        private void tvConfig_DragLeave(object sender, EventArgs e)
        {
            tvConfig.Refresh();
        }

        private void tvConfig_ItemDrag(object sender, ItemDragEventArgs e)
        {
            DoDragDrop(e.Item, DragDropEffects.Move);
        }

        private void DrawPlaceholder(TreeNode node, int pos)
        {
            var graphics = tvConfig.CreateGraphics();
            var bounds = node.Bounds;
            if (pos == 0)
            {
                var midY = bounds.Height / 2 + bounds.Y;
                var x = bounds.Right + 6;
                
                graphics.FillPolygon
                (
                    System.Drawing.Brushes.Black,
                    new Point[]
                    {
                        new Point(x, midY + 4),
					    new Point(x - 4, midY),
					    new Point(x, midY - 4)
                    }
                );
            }
            else // below
            {
                graphics.DrawLine(Pens.Black,
                    new Point(bounds.Left, pos < 0 ? bounds.Top : bounds.Bottom),
                    new Point(bounds.Right, pos < 0 ? bounds.Top : bounds.Bottom));
            }
        }        
    }

    public static class Extensions
    {
        public static bool IsChildOf(this TreeNode child, TreeNode parent)
        {
            if (child.Parent == parent || child == parent)
            {
                return true;
            }
            else if (child.Parent == null)
            {
                return false;
            }
            else
            {
                return child.Parent.IsChildOf(parent);
            }
        }
    }
}
